@@ -16,6 +16,7 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
+#include <QFileDialog>
 #include <QColorDialog>
 #include <QClipboard>
 #include <QTimer>
@@ -34,13 +35,48 @@ PwmdDialog::PwmdDialog(const QString &s, const QString &f, const QString &name,
 	QWidget *p) : QDialog(p)
 {
     ui.setupUi(this);
-    QSettings cfg("pwmdEdit");
+    QSettings cfg("pwmdDialog");
+    ipVersion = cfg.value("ipVersion", 0).toInt();
+
+    switch (ipVersion) {
+	case 0:
+	    ui.rb_ipany->setChecked(true);
+	    break;
+	case 1:
+	    ui.rb_ipv4->setChecked(true);
+	    break;
+	case 2:
+	    ui.rb_ipv6->setChecked(true);
+	    break;
+    }
+
+    connect(ui.rb_ipany, SIGNAL(toggled(bool)), SLOT(slotIpVersionToggled(bool)));
+    connect(ui.rb_ipv4, SIGNAL(toggled(bool)), SLOT(slotIpVersionToggled(bool)));
+    connect(ui.rb_ipv6, SIGNAL(toggled(bool)), SLOT(slotIpVersionToggled(bool)));
+    ui.le_socket->setText(cfg.value("defaultSocket", "").toString());
+    ui.le_hostname->setText(cfg.value("hostname", "").toString());
+    ui.sp_port->setValue(cfg.value("port", 22).toInt());
+    ui.le_username->setText(cfg.value("username", "").toString());
+    ui.le_identity->setText(cfg.value("identity", "").toString());
+    ui.le_knownhosts->setText(cfg.value("knownhosts", "").toString());
+    ui.ck_remote->setChecked(cfg.value("remote", false).toBool());
+
+    if (ui.le_knownhosts->text().isEmpty())
+	ui.le_knownhosts->setText(QDir::home().path() + "/.ssh/known_hosts");
+
+    if (ui.le_identity->text().isEmpty())
+	ui.le_identity->setText(QDir::home().path() + "/.ssh/pwmd");
+
+    if (ui.le_username->text().isEmpty())
+	ui.le_username->setText(getlogin());
 
     if (s.isEmpty())
-	sock = cfg.value("defaultSocket", "").toString();
+	sock = updateSocket();
     else
 	sock = s;
 
+    connect(ui.ck_remote, SIGNAL(toggled(bool)), SLOT(slotRemoteToggled(bool)));
+
     ui.ck_reopenLastFile->setChecked(cfg.value("reopenLastFile", false).toBool());
     targetColor = cfg.value("targetColor", "DarkOrange").value<QColor>();
     invalidTargetColor = cfg.value("invalidTargetColor", "Red").value<QColor>();
@@ -57,10 +93,13 @@ PwmdDialog::PwmdDialog(const QString &s, const QString &f, const QString &name,
     if (ui.ck_geometry->isChecked())
 	setGeometry(cfg.value("geometry").toRect());
 
-    pwm = new PwmdPrivate(s, filename(), name, this);
+    pwm = new PwmdPrivate(sock, filename(), name, this);
     connect(pwm, SIGNAL(stateChanged(Pwmd::ConnectionState)), this, SLOT(slotConnectionStateChanged(Pwmd::ConnectionState)));
     connect(pwm, SIGNAL(connectionError(unsigned)), this, SLOT(slotConnectionError(unsigned)));
     pwm->setStatusCallback(statusCallback, this);
+#if LIBPWMD_VERSION >= 0x603
+    pwm->setKnownhostCallback(knownhostCallback, NULL);
+#endif
     dndOperation = false;
     previousElement = NULL;
     previousRightPage = NULL;
@@ -69,7 +108,11 @@ PwmdDialog::PwmdDialog(const QString &s, const QString &f, const QString &name,
     selectedFilename = QString::null;
     iterationsOrig = 0;
     ui.l_pwmdDialogVersion->setText(VERSION);
-    ui.l_libpwmdVersion->setText(QString("0x%1").arg(LIBPWMD_VERSION, 0, 16));
+#if LIBPWMD_VERSION >= 0x603
+    ui.l_libpwmdVersion->setText(pwmd_version());
+#else
+    ui.l_libpwmdVersion->setText(QString::number(LIBPWMD_VERSION, 16));
+#endif
     ui.l_pwmdVersion->setText("-");
     ui.elementTree->setParent(this);
     ui.elementTree->header()->setResizeMode(0, QHeaderView::ResizeToContents);
@@ -193,6 +236,8 @@ PwmdDialog::PwmdDialog(const QString &s, const QString &f, const QString &name,
     connect(ui.pb_options, SIGNAL(clicked()), SLOT(slotEditOptions()));
     connect(ui.tb_optionsOk, SIGNAL(clicked()), SLOT(slotOptionsOk()));
     connect(ui.tb_optionsCancel, SIGNAL(clicked()), SLOT(slotOptionsCancel()));
+    connect(ui.pb_identity, SIGNAL(clicked()), SLOT(slotChooseIdentityFile()));
+    connect(ui.pb_knownhosts, SIGNAL(clicked()), SLOT(slotChooseKnownhostsFile()));
 
     QPushButton *pb = ui.buttonBox->button(QDialogButtonBox::Apply);
     connect(pb, SIGNAL(clicked()), SLOT(slotSaveDocument()));
@@ -206,7 +251,7 @@ PwmdDialog::PwmdDialog(const QString &s, const QString &f, const QString &name,
 PwmdDialog::~PwmdDialog()
 {
     delete pwm;
-    QSettings cfg("pwmdEdit");
+    QSettings cfg("pwmdDialog");
 
     cfg.setValue("defaultSocket", ui.le_socket->text());
     cfg.setValue("reopenLastFile", ui.ck_reopenLastFile->isChecked());
@@ -217,6 +262,88 @@ PwmdDialog::~PwmdDialog()
     cfg.setValue("showContent", ui.ck_showContent->isChecked());
     cfg.setValue("rememberGeometry", ui.ck_geometry->isChecked());
     cfg.setValue("geometry", geometry());
+    cfg.setValue("remote", ui.ck_remote->isChecked());
+    cfg.setValue("hostname", ui.le_hostname->text());
+    cfg.setValue("port", ui.sp_port->value());
+    cfg.setValue("username", ui.le_username->text());
+    cfg.setValue("identity", ui.le_identity->text());
+    cfg.setValue("knownhosts", ui.le_knownhosts->text());
+    cfg.setValue("ipVersion", ipVersion);
+}
+
+void PwmdDialog::slotIpVersionToggled(bool b)
+{
+    (void)b;
+
+    if (ui.rb_ipany->isChecked())
+	ipVersion = 0;
+    else if (ui.rb_ipv4->isChecked())
+	ipVersion = 1;
+    else if (ui.rb_ipv6->isChecked())
+	ipVersion = 2;
+}
+
+void PwmdDialog::slotChooseIdentityFile()
+{
+    QString f = QFileDialog::getOpenFileName(this,
+	    tr("SSH Identity file"), QDir::home().path());
+
+    if (f.isEmpty())
+	return;
+
+    ui.le_identity->setText(f);
+}
+
+void PwmdDialog::slotChooseKnownhostsFile()
+{
+    QString f = QFileDialog::getOpenFileName(this,
+	    tr("SSH knownhosts file"), QDir::home().path());
+
+    if (f.isEmpty())
+	return;
+
+    ui.le_knownhosts->setText(f);
+}
+
+void PwmdDialog::slotRemoteToggled(bool b)
+{
+    ui.le_socket->setEnabled(!b);
+}
+
+QString PwmdDialog::updateSocket()
+{
+    if (ui.ck_remote->isChecked()) {
+	QString s = "ssh";
+	
+	switch (ipVersion) {
+	    case 0:
+		break;
+	    case 1:
+		s.append("4");
+		break;
+	    case 2:
+		s.append("6");
+		break;
+	}
+	
+	s.append("://");
+
+	if (!ui.le_username->text().isEmpty()) {
+	    s.append(ui.le_username->text());
+	    s.append('@');
+	}
+
+	s.append(ui.le_hostname->text());
+	s.append(':');
+	s.append(QString("%1").arg(ui.sp_port->value()));
+	s.append(',');
+	s.append(ui.le_identity->text());
+	s.append(',');
+	s.append(ui.le_knownhosts->text());
+	return s;
+    }
+
+    return ui.le_socket->text();
 }
 
 void PwmdDialog::slotConnectionError(unsigned rc)
@@ -587,7 +714,7 @@ void PwmdDialog::slotExpandAll()
 
 void PwmdDialog::slotEditOptions()
 {
-    ui.le_socket->setText(sock);
+    remote = ui.ck_remote->isChecked();
 
     if (state != Pwmd::Init) {
 	gpg_error_t rc;
@@ -604,6 +731,7 @@ void PwmdDialog::slotEditOptions()
 	    ui.l_pwmdVersion->setText(data);
     }
 
+    slotRemoteToggled(remote);
     ui.tb_targetColor->setPalette(QPalette(targetColor));
     ui.tb_invalidTargetColor->setPalette(QPalette(invalidTargetColor));
     ui.tb_hilightColor->setPalette(QPalette(hilightColor));
@@ -631,9 +759,9 @@ void PwmdDialog::slotOptionsOk()
     ui.stackedWidgetRight->setEnabled(selectedElement != NULL);
     ui.stackedWidgetLeft->setCurrentWidget(ui.elementPage);
 
-    if (sock != ui.le_socket->text()) {
+    if (sock != updateSocket()) {
 	if (doFinished(false)) {
-	    sock = ui.le_socket->text();
+	    sock = updateSocket();
 	    pwm->reset();
 	    pwm->setSocket(sock);
 	    slotConnectSocket();
@@ -646,6 +774,8 @@ void PwmdDialog::slotOptionsOk()
 
 void PwmdDialog::slotOptionsCancel()
 {
+    slotRemoteToggled(remote);
+    ui.ck_remote->setChecked(remote);
     ui.bottomFrame->setHidden(false);
     ui.stackedWidgetRight->setHidden(false);
     ui.stackedWidgetLeft->setCurrentWidget(ui.elementPage);
@@ -832,6 +962,7 @@ void PwmdDialog::slotPasswordContentChanged(const QString &str)
     bool b = ui.le_password->isModified();
 
     ui.tb_passwordUpdate->setEnabled(b);
+    setModified();
 }
 
 void PwmdDialog::tabify(QPlainTextEdit *w)
@@ -866,6 +997,8 @@ void PwmdDialog::slotContentChanged()
 	if (selectedAttribute->text() == "target")
 	    tabify(ui.textEdit);
     }
+
+    setModified();
 }
 
 void PwmdDialog::setBusy(bool b)
@@ -916,6 +1049,9 @@ void PwmdDialog::showError(gpg_error_t rc)
     m.setInformativeText(QString("ERR %1: %2").arg(rc).arg(pwmd_strerror(rc)));
 
     switch (rc) {
+	case EPWMD_FILE_MODIFIED:
+	    m.setDetailedText(tr("The currently open data file has been externally modified; either by another client or by some other means. Reopen the file to fix this error."));
+	    break;
 	case GPG_ERR_LOCKED:
 	    m.setDetailedText(tr("Another client has the same data file opened and has requested that other clients must wait until it has finished."));
 	    break;
@@ -1043,6 +1179,7 @@ void PwmdDialog::setConnected(bool b)
     ui.pb_open->setFocus();
     ui.progressBar->setValue(0);
     ui.progressBar->setFormat("");
+    ui.tb_passwordUpdate->setEnabled(false);
 
     QPushButton *pb = ui.buttonBox->button(QDialogButtonBox::Ok);
     pb->setEnabled(b);
@@ -1091,21 +1228,48 @@ gpg_error_t PwmdDialog::refreshElementTree(bool target)
 
 void PwmdDialog::slotConnectSocket()
 {
+    setBusy();
     gpg_error_t rc = pwm->connect();
 
     if (rc) {
+	setBusy(false);
 	showError(rc);
 	return;
     }
 
     if (filename().isEmpty()) {
 	slotOpenFile();
+	setBusy(false);
 	return;
     }
 
+    setBusy(false);
     rc = refreshElementTree();
 }
 
+#if LIBPWMD_VERSION >= 0x603
+gpg_error_t PwmdDialog::knownhostCallback(void *data, const char *host,
+	const char *key, size_t len)
+{
+    (void)key;
+    (void)len;
+    (void)data;
+    QMessageBox msgBox;
+    msgBox.setText(QString(tr("While attempting an SSH connection to %1, there was a problem verifying it's host key against the known and trusted hosts file because it's host key was not found.\n\nWould you like to treat this connection as trusted for this and future connections by adding %2's host key to the known hosts file?")).arg(host).arg(host));
+
+    msgBox.setIcon(QMessageBox::Question);
+    msgBox.setInformativeText(tr("Accept this host key as trusted?"));
+    msgBox.setStandardButtons(QMessageBox::Save | QMessageBox::Cancel);
+    msgBox.setDefaultButton(QMessageBox::Save);
+    int n = msgBox.exec();
+
+    if (n == QMessageBox::Cancel)
+	return GPG_ERR_CANCELED;
+
+    return 0;
+}
+#endif
+
 int PwmdDialog::statusCallback(void *data, const char *line)
 {
     PwmdDialog *w = static_cast<PwmdDialog *>(data);
@@ -1220,7 +1384,7 @@ bool PwmdDialog::verifyNewContent()
 	}
     }
     else {
-	ui.le_password->setModified(false);
+	setPasswordContent("");
 	ui.tb_passwordUpdate->setEnabled(false);
 	ui.pb_update->setEnabled(false);
 	ui.textEdit->document()->setModified(false);
