@@ -20,6 +20,7 @@
 #define PWMDEDIT_H
 
 #include <QSystemTrayIcon>
+#include <QMenu>
 #include "pwmdDialog.h"
 #include "pwmdShortcutDialog.h"
 
@@ -39,10 +40,13 @@ class PwmdEdit : public QApplication
 	void slotClearClipboard();
 
     private:
+	void refreshShortcuts();
+
 	QString filename;
 	QString socket;
 	QSystemTrayIcon *sti;
 	int clipboardTimeout;
+	QMenu *shortcutMenu;
 };
 
 #endif
